/*
** XSwarm
** ======
**
** Purpose:	Display a swarm of bees chasing a wasp.
**
** Features:	uses only integer math
**		can be put in the root window
**		uses non-blocking delays to decrease cpu usage
**		has no redeeming social value
**
** Comments:	I used psychoII (I assume it's a derivative of ico.) as an
**		an example of how to open up a window and draw some lines.
**		Most of this program is now totally rewritten by me, but the
**		initial window manager hints stuff is left over from the older
**		program.
**
**		The program meltdown was handy for learning how to take over
**		the screen.
**
**		Any comments, enhances, or fixes are welcome!
**
** Created:	by Jeff Butterworth on 7/11/90
**		butterwo@cs.unc.edu
**
** Updated:	by Jeff Butterworth on 1/4/91
**
*/

/* These are needed for the nap function. */
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ioctl.h>

/* Standard Includes */
#include <stdio.h>

#include "vfdlib.h"
/* Includes for this project. */
#include "xswarm.h"

void Animate(void);
void nap(long sec, long usec);
void *Safe_Malloc(int bytes);


main(int argc, char **argv)
{
    Animate();
}



/* Animate()								*/
/* Move the swarm around.						*/

void
Animate(void)
{
    register int	b;		/* bee index */
    int			*x, *y;		/* bee positions x[time][bee#] */
    int			*xv, *yv;	/* bee velocities xv[bee#] */
    int			wx[3], wy[3];	/* wasp positions */
    int			wxv, wyv;	/* wasp velocity */
    int			dx,dy,distance;
    int			init = TRUE;
    
    char *              display = NULL;
    int                 displayFD = 0;
    
    int bx1, by1, bx2, by2, ex1, ey1, ex2, ey2;  

    /* Open display device */
    displayFD = open("/dev/display", O_RDWR);
    if (displayFD == -1) return;
    
    #define BLIT(fd) ioctl((fd), _IO('d',0))
    #define DISPLAY(fd, x) ioctl((fd), _IOW('d', 1, int), (x))

    display = (char *) mmap(0, 4096, PROT_READ | PROT_WRITE, MAP_SHARED, displayFD, 0);
    DISPLAY(displayFD, 1);
    BLIT(displayFD);


        /* Set up window size. */
	winW = VFD_WIDTH;
	winH = VFD_HEIGHT;
	winX = 0;
	winY = 0;
        vfdlib_clear(display, VFDSHADE_BLACK);

    /* Allocate memory. */
    x = (int *) Safe_Malloc(sizeof(int) * bees * times);
    y = (int *) Safe_Malloc(sizeof(int) * bees * times);
    xv = (int *) Safe_Malloc(sizeof(int) * bees);
    yv = (int *) Safe_Malloc(sizeof(int) * bees);

 
    /* Initialize point positions, velocities, etc. */

    /* wasp */
    {
	wx[0] = BORDER + random() % (winW - 2*BORDER);
	wy[0] = BORDER + random() % (winH - 2*BORDER);
    }
    wx[1] = wx[0];
    wy[1] = wy[0];
    wxv = 0;
    wyv = 0;

    /* bees */
    for (b = 0 ; b < bees ; b++)
    {
	X(0,b) = random() % winW;
	X(1,b) = X(0,b);
	Y(0,b) = random() % winH;
	Y(1,b) = Y(0,b);
	xv[b] = RAND(7);
	yv[b] = RAND(7);
    }

    /* Seemingly endless loop. */
    while (TRUE)
    {
	/* <=- Wasp -=> */
	/* Age the position arrays. */
	wx[2] = wx[1];
	wx[1] = wx[0];
	wy[2] = wy[1];
	wy[1] = wy[0];

	{
	    /* Accelerate */
	    wxv += RAND(wasp_acc);
	    wyv += RAND(wasp_acc);

	    /* Speed Limit Checks */
	    if (wxv > wasp_vel) wxv = wasp_vel;
	    if (wxv < -wasp_vel) wxv = -wasp_vel;
	    if (wyv > wasp_vel) wyv = wasp_vel;
	    if (wyv < -wasp_vel) wyv = -wasp_vel;

	    /* Move */
	    wx[0] = wx[1] + wxv;
	    wy[0] = wy[1] + wyv;

	    /* Bounce Checks */
	    if ((wx[0] < border) || (wx[0] > winW-border-1))
	    {
		wxv = -wxv;
		wx[0] += wxv<<1;
	    }
	    if ((wy[0] < border) || (wy[0] > winH-border-1))
	    {
		wyv = -wyv;
		wy[0] += wyv<<1;
	    }
	}

	/* Don't let things settle down. */
	if (bees > 0) /* Avoid division by 0. */
	{
	    xv[random() % bees] += RAND(3);
	    yv[random() % bees] += RAND(3);
	}

	/* <=- Bees -=> */
	for (b = 0 ; b < bees ; b++)
	{
	    /* Age the arrays. */
	    X(2,b) = X(1,b);
	    X(1,b) = X(0,b);
	    Y(2,b) = Y(1,b);
	    Y(1,b) = Y(0,b);

	    /* Accelerate */
	    dx = wx[1] - X(1,b);
	    dy = wy[1] - Y(1,b);
	    distance = abs(dx)+abs(dy); /* approximation */
	    if (distance == 0) distance = 1;
	    xv[b] += (dx*bee_acc)/distance;
	    yv[b] += (dy*bee_acc)/distance;

	    /* Speed Limit Checks */
	    if (xv[b] > bee_vel) xv[b] = bee_vel;
	    if (xv[b] < -bee_vel) xv[b] = -bee_vel;
	    if (yv[b] > bee_vel) yv[b] = bee_vel;
	    if (yv[b] < -bee_vel) yv[b] = -bee_vel;

	    /* Move */
	    X(0,b) = X(1,b) + xv[b];
	    Y(0,b) = Y(1,b) + yv[b];

	/* Bees */

	    /* Fill the segment lists. */
	    bx1 = X(0,b);
	    by1 = Y(0,b);
	    bx2 = X(1,b);
	    by2 = Y(1,b);
	    ex1 = X(1,b);
	    ey1 = Y(1,b);
	    ex2 = X(2,b);
	    ey2 = Y(2,b);
	    
	    if (init == FALSE)
		vfdlib_drawLineClipped(display, ex1, ey1, ex2, ey2, VFDSHADE_BLACK);
	    else
	        init = FALSE;
	    vfdlib_drawLineClipped(display, bx1, by1, bx2, by2, VFDSHADE_MEDIUM);

	}

	/* Erase previous, draw current, sync for smoothness. */


	/* Wasp */
	    if (init == FALSE)
		vfdlib_drawLineClipped(display, wx[1], wy[1], wx[2], wy[2], VFDSHADE_BLACK);
	   vfdlib_drawLineClipped(display, wx[0], wy[0], wx[1], wy[1], VFDSHADE_BRIGHT);

          BLIT(displayFD);

	/* Check for events. */
	if (0)//XPending(display.dpy))
	{
	   // XNextEvent(display.dpy, &xev);
	   // HandleEvent(&xev);

	    /* Make sure the wasp is within the window. */
	    if (wx[0] < border) wx[0] = border;
	    if (wx[0] > winW-border-1) wx[0] = winW-border-1;
	    if (wy[0] < border) wy[0] = border;
	    if (wy[0] > winH-border-1) wy[0] = winH-border-1;
	}
	
	/* Clean up and shut down. */
	if (stop)
	{
	    stop = FALSE; /* reset the "stop" variable */

	    vfdlib_clear(display, VFDSHADE_BLACK);

	    
	    /* Free Memory */
	    free(x);
	    free(y);
	    free(xv);
	    free(yv);

	    return;
	}

	/* Delay so we don't use all of the cpu time. */
	if (delay != 0) nap(0,delay);
    }
}


/*  Put the process to sleep for a while.  */
void nap(long sec, long usec)
{
    static struct timeval tv;

    tv.tv_sec = sec;
    tv.tv_usec = usec;
    select(0, 0, 0, 0, &tv);
}



/* Check the return code of malloc(). */
void *
Safe_Malloc(int bytes)
{
    void *pointer;

    pointer = (void *) malloc(bytes);
    if (NULL == pointer)
    {
	fprintf(stderr, "Error allocating %d bytes.\n", bytes);
	exit(-1);
    }

    return(pointer);
}


